from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from .event import Event


class EventDispatcher(object):
    def __init__(self):  # type: () -> None
        self._listeners = {}
        self._sorted = {}

    def dispatch(self, event_name, event=None):  # type: (str, Event) -> Event
        if event is None:
            event = Event()

        listeners = self.get_listeners(event_name)

        if listeners:
            self._do_dispatch(listeners, event_name, event)

        return event

    def get_listeners(
        self, event_name=None
    ):  # type: (str) -> Union[List[Callable], Dict[str, Callable]]
        if event_name is not None:
            if event_name not in self._listeners:
                return []

            if event_name not in self._sorted:
                self._sort_listeners(event_name)

            return self._sorted[event_name]

        for event_name, event_listeners in self._listeners.items():
            if event_name not in self._sorted:
                self._sort_listeners(event_name)

        return self._sorted

    def get_listener_priority(
        self, event_name, listener
    ):  # type: (str, Callable) -> Optional[int]
        if event_name not in self._listeners:
            return

        for priority, listeners in self._listeners[event_name].items():
            for v in listeners:
                if v == listener:
                    return priority

    def has_listeners(self, event_name=None):  # type: (Optional[str]) -> bool
        if event_name is not None:
            if event_name not in self._listeners:
                return False

            return len(self._listeners[event_name]) > 0

        for event_listeners in self._listeners.values():
            if event_listeners:
                return True

        return False

    def add_listener(
        self, event_name, listener, priority=0
    ):  # type: (str, Callable, int) -> None
        if event_name not in self._listeners:
            self._listeners[event_name] = {}

        if priority not in self._listeners[event_name]:
            self._listeners[event_name][priority] = []

        self._listeners[event_name][priority].append(listener)

        if event_name in self._sorted:
            del self._sorted[event_name]

    def _do_dispatch(
        self, listeners, event_name, event
    ):  # type: (List[Callable], str, Event) -> None
        for listener in listeners:
            if event.is_propagation_stopped():
                break

            listener(event, event_name, self)

    def _sort_listeners(self, event_name):  # type: (str) -> None
        """
        Sorts the internal list of listeners for the given event by priority.
        """
        self._sorted[event_name] = []

        for priority, listeners in sorted(
            self._listeners[event_name].items(), key=lambda t: -t[0]
        ):
            for listener in listeners:
                self._sorted[event_name].append(listener)



from typing import Optional


class InputStream(object):
    """
    The console input stream.
    """

    def read(self, length):  # type: (int) -> str
        """
        Reads the given amount of characters from the stream.
        """
        raise NotImplementedError()

    def read_line(self, length=None):  # type: (Optional[int]) -> str
        """
        Reads a line from the stream.
        """
        raise NotImplementedError()

    def close(self):  # type: () -> None
        """
        Closes the stream
        """
        raise NotImplementedError()

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the stream is closed or not
        """
        raise NotImplementedError()



from typing import Optional

from pastel import Pastel

from clikit.adapter.style_converter import StyleConverter
from clikit.api.formatter import Formatter
from clikit.api.formatter import Style
from clikit.api.formatter import StyleSet

from .default_style_set import DefaultStyleSet


class PlainFormatter(Formatter):
    """
    A formatter that removes all format tags.
    """

    def __init__(self, style_set=None):  # type: (StyleSet) -> None
        self._formatter = Pastel(False)

        if style_set is None:
            style_set = DefaultStyleSet()

        for tag, style in style_set.styles.items():
            pastel_style = StyleConverter.convert(style)

            self._formatter.add_style(
                tag,
                pastel_style.foreground,
                pastel_style.background,
                pastel_style.options,
            )

    def format(self, string, style=None):  # type: (str, Optional[Style]) -> str
        return self._formatter.colorize(string)

    def remove_format(self, string):  # type: (str) -> str
        return self._formatter.colorize(string)

    def disable_ansi(self):  # type: () -> bool
        return True

    def force_ansi(self):  # type: () -> bool
        return False

    def add_style(self, style):  # type: (Style) -> None
        pastel_style = StyleConverter.convert(style)

        self._formatter.add_style(
            style.tag,
            pastel_style.foreground,
            pastel_style.background,
            pastel_style.options,
        )



from clikit.api.exceptions import CliKitException
from clikit.utils.command import find_similar_command_names


class CannotResolveCommandException(RuntimeError, CliKitException):
    @classmethod
    def name_not_found(cls, name, commands):
        message = 'The command "{}" is not defined.'.format(name)

        suggested_names = find_similar_command_names(name, commands)

        if suggested_names:
            if len(suggested_names) == 1:
                message += "\n\nDid you mean this?\n    "
            else:
                message += "\n\nDid you mean one of these?\n    "

            message += "\n    ".join(suggested_names)

        return cls(message)

    @classmethod
    def no_default_command(cls):
        return cls("No default command is defined.")



from .border_util import BorderUtil
from .cell_wrapper import CellWrapper
from .choice_question import ChoiceQuestion
from .confirmation_question import ConfirmationQuestion
from .empty_line import EmptyLine
from .exception_trace import ExceptionTrace
from .labeled_paragraph import LabeledParagraph
from .name_version import NameVersion
from .paragraph import Paragraph
from .progress_bar import ProgressBar
from .progress_indicator import ProgressIndicator
from .question import Question
from .table import Table



from typing import Optional

from clikit.api.io import IO
from clikit.api.io import Input
from clikit.api.io import Output
from clikit.formatter import AnsiFormatter
from clikit.formatter import PlainFormatter
from clikit.utils.terminal import Terminal
from clikit.ui.rectangle import Rectangle

from .input_stream import StandardInputStream
from .output_stream import ErrorOutputStream
from .output_stream import StandardOutputStream


class ConsoleIO(IO):
    """
    An I/O that reads from/prints to the console.
    """

    def __init__(
        self, input=None, output=None, error_output=None
    ):  # type: (Optional[Input], Optional[Output], Optional[Output]) -> None
        if input is None:
            input_stream = StandardInputStream()
            input = Input(input_stream)

        if output is None:
            output_stream = StandardOutputStream()
            if output_stream.supports_ansi():
                formatter = AnsiFormatter()
            else:
                formatter = PlainFormatter()

            output = Output(output_stream, formatter)

        if error_output is None:
            error_stream = ErrorOutputStream()
            if error_stream.supports_ansi():
                formatter = AnsiFormatter()
            else:
                formatter = PlainFormatter()

            error_output = Output(error_stream, formatter)

        super(ConsoleIO, self).__init__(input, output, error_output)

    def get_default_terminal_dimensions(self):  # type: () -> Rectangle
        terminal = Terminal()

        return Rectangle(terminal.width, terminal.height)



from typing import Optional

from .input_stream import InputStream


class Input(object):
    """
    The console input.

    This class wraps an input stream and adds convenience functionality for
    reading that stream.
    """

    def __init__(self, stream):  # type: (InputStream) -> None
        self._stream = stream
        self._interactive = True

    def read(self, length, default=None):  # type: (int, Optional[str]) -> str
        """
        Reads the given amount of characters from the input stream.

        :raises: IOException
        """
        if not self._interactive:
            return default

        return self._stream.read(length)

    def read_line(
        self, length=None, default=None
    ):  # type: (Optional[int], Optional[str]) -> str
        """
        Reads a line from the input stream.

        :raises: IOException
        """
        if not self._interactive:
            return default

        return self._stream.read_line(length=length)

    def close(self):  # type: () -> None
        """
        Closes the input.
        """
        self._stream.close()

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the input is closed.
        """
        return self._stream.is_closed()

    def set_stream(self, stream):  # type: (InputStream) -> None
        """
        Sets the underlying stream.
        """
        self._stream = stream

    @property
    def stream(self):  # type: () -> InputStream
        return self._stream

    def set_interactive(self, interactive):  # type: (bool) -> None
        """
        Enables or disables interaction with the user.
        """
        self._interactive = interactive

    def is_interactive(self):  # type: () -> bool
        """
        Returns whether the user may be asked for input.
        """
        return self._interactive



# -*- coding: utf-8 -*-
import ast
import inspect
import io
import keyword
import os
import re
import sys
import tokenize
import traceback

from clikit.api.io import IO
from clikit.formatter.plain_formatter import PlainFormatter
from clikit.utils._compat import PY2
from clikit.utils._compat import PY36
from clikit.utils._compat import decode
from clikit.utils._compat import encode


class Highlighter(object):

    TOKEN_DEFAULT = "token_default"
    TOKEN_COMMENT = "token_comment"
    TOKEN_STRING = "token_string"
    TOKEN_NUMBER = "token_number"
    TOKEN_KEYWORD = "token_keyword"
    TOKEN_BUILTIN = "token_builtin"
    TOKEN_OP = "token_op"
    LINE_MARKER = "line_marker"
    LINE_NUMBER = "line_number"

    DEFAULT_THEME = {
        TOKEN_STRING: "fg=yellow;options=bold",
        TOKEN_NUMBER: "fg=blue;options=bold",
        TOKEN_COMMENT: "fg=default;options=dark,italic",
        TOKEN_KEYWORD: "fg=magenta;options=bold",
        TOKEN_BUILTIN: "fg=default;options=bold",
        TOKEN_DEFAULT: "fg=default",
        TOKEN_OP: "fg=default;options=dark",
        LINE_MARKER: "fg=red;options=bold",
        LINE_NUMBER: "fg=default;options=dark",
    }

    KEYWORDS = set(keyword.kwlist)
    BUILTINS = set(
        __builtins__.keys() if type(__builtins__) is dict else dir(__builtins__)
    )

    def __init__(self):
        self._theme = self.DEFAULT_THEME.copy()

    def code_snippet(self, source, line, lines_before=2, lines_after=2):
        token_lines = self.highlighted_lines(source)
        token_lines = self.line_numbers(token_lines, line)

        offset = line - lines_before - 1
        offset = max(offset, 0)
        length = lines_after + lines_before + 1
        token_lines = token_lines[offset : offset + length]

        return token_lines

    def highlighted_lines(self, source):
        source = source.replace("\r\n", "\n").replace("\r", "\n")

        return self.split_to_lines(source)

    def split_to_lines(self, source):
        lines = []
        current_line = 1
        current_col = 0
        buffer = ""
        current_type = None
        source_io = io.BytesIO(encode(source))
        formatter = PlainFormatter()

        def readline():
            return encode(formatter.remove_format(decode(source_io.readline())))

        tokens = tokenize.tokenize(readline)
        line = ""
        for token_info in tokens:
            token_type, token_string, start, end, _ = token_info
            lineno = start[0]
            if lineno == 0:
                # Encoding line
                continue

            if token_type == tokenize.ENDMARKER:
                # End of source
                lines.append(line)
                break

            if lineno > current_line:
                diff = lineno - current_line
                if diff > 1:
                    lines += [""] * (diff - 1)

                line += "<{}>{}</>".format(
                    self._theme[current_type], buffer.rstrip("\n")
                )

                # New line
                lines.append(line)
                line = ""
                current_line = lineno
                current_col = 0
                buffer = ""

            if token_string in self.KEYWORDS:
                new_type = self.TOKEN_KEYWORD
            elif token_string in self.BUILTINS or token_string == "self":
                new_type = self.TOKEN_BUILTIN
            elif token_type == tokenize.STRING:
                new_type = self.TOKEN_STRING
            elif token_type == tokenize.NUMBER:
                new_type = self.TOKEN_NUMBER
            elif token_type == tokenize.COMMENT:
                new_type = self.TOKEN_COMMENT
            elif token_type == tokenize.OP:
                new_type = self.TOKEN_OP
            elif token_type == tokenize.NEWLINE:
                continue
            else:
                new_type = self.TOKEN_DEFAULT

            if current_type is None:
                current_type = new_type

            if start[1] > current_col:
                buffer += token_info.line[current_col : start[1]]

            if current_type != new_type:
                line += "<{}>{}</>".format(self._theme[current_type], buffer)
                buffer = ""
                current_type = new_type

            if lineno < end[0]:
                # The token spans multiple lines
                lines.append(line)
                token_lines = token_string.split("\n")
                for l in token_lines[1:-1]:
                    lines.append("<{}>{}</>".format(self._theme[current_type], l))

                current_line = end[0]
                buffer = token_lines[-1][: end[1]]
                line = ""
                continue

            buffer += token_string
            current_col = end[1]
            current_line = lineno

        return lines

    def line_numbers(self, lines, mark_line=None):
        max_line_length = len(str(len(lines)))

        snippet_lines = []
        marker = "  <{}>→</> ".format(self._theme[self.LINE_MARKER])
        no_marker = "    "
        for i, line in enumerate(lines):
            if mark_line is not None:
                if mark_line == i + 1:
                    snippet = marker
                else:
                    snippet = no_marker

            line_number = "{:>{}}".format(i + 1, max_line_length)
            snippet += "<{}>{}</><{}>│</> {}".format(
                "fg=default;options=bold"
                if mark_line == i + 1
                else self._theme[self.LINE_NUMBER],
                line_number,
                self._theme[self.LINE_NUMBER],
                line,
            )
            snippet_lines.append(snippet)

        return snippet_lines


class ExceptionTrace(object):
    """
    Renders the trace of an exception.
    """

    THEME = {
        "comment": "<fg=black;options=bold>",
        "keyword": "<fg=yellow>",
        "builtin": "<fg=blue>",
        "literal": "<fg=magenta>",
    }

    AST_ELEMENTS = {
        "builtins": __builtins__.keys()
        if type(__builtins__) is dict
        else dir(__builtins__),
        "keywords": [
            getattr(ast, cls)
            for cls in dir(ast)
            if keyword.iskeyword(cls.lower())
            and inspect.isclass(getattr(ast, cls))
            and issubclass(getattr(ast, cls), ast.AST)
        ],
    }

    _FRAME_SNIPPET_CACHE = {}

    def __init__(
        self, exception, solution_provider_repository=None
    ):  # type: (Exception, ...) -> None
        self._exception = exception
        self._solution_provider_repository = solution_provider_repository
        self._exc_info = sys.exc_info()
        self._higlighter = Highlighter()
        self._ignore = None

    def ignore_files_in(self, ignore):  # type: (str) -> ExceptionTrace
        self._ignore = ignore

        return self

    def render(self, io, simple=False):  # type: (IO, bool) -> None
        if simple:
            io.write_line("<error>{}</error>".format(str(self._exception)))
            return

        if not PY36:
            return self._render_legacy(io)

        return self._render_exception(io, self._exception)

    def _render_legacy(self, io):
        if hasattr(self._exception, "__traceback__"):
            tb = self._exception.__traceback__
        else:
            tb = self._exc_info[2]

        title = "\n<error>{}</error>\n\n<b>{}</b>".format(
            self._exception.__class__.__name__, str(self._exception)
        )

        io.write_line(title)

        if io.is_verbose():
            io.write_line("")
            self._render_traceback(io, tb)

    def _render_exception(self, io, exception):
        from crashtest.inspector import Inspector

        inspector = Inspector(exception)
        if not inspector.frames:
            return

        self._render_trace(io, inspector.frames)

        self._render_line(
            io, "<error>{}</error>".format(inspector.exception_name), True
        )
        io.write_line("")
        exception_message = io.remove_format(inspector.exception_message).replace(
            "\n", "\n  "
        )
        self._render_line(io, "<b>{}</b>".format(exception_message))

        current_frame = inspector.frames[-1]
        self._render_snippet(io, current_frame)

        self._render_solution(io, inspector)

    def _render_snippet(self, io, frame):
        self._render_line(
            io,
            "at <fg=green>{}</>:<b>{}</b> in <fg=cyan>{}</>".format(
                self._get_relative_file_path(frame.filename),
                frame.lineno,
                frame.function,
            ),
            True,
        )

        code_lines = self._higlighter.code_snippet(
            frame.file_content, frame.lineno, 4, 4
        )

        for code_line in code_lines:
            self._render_line(io, code_line)

    def _render_solution(self, io, inspector):
        if self._solution_provider_repository is None:
            return

        solutions = self._solution_provider_repository.get_solutions_for_exception(
            inspector.exception
        )
        for solution in solutions:
            title = solution.solution_title
            description = solution.solution_description
            links = solution.documentation_links

            description = description.replace("\n", "\n    ").strip(" ")

            self._render_line(
                io,
                "<fg=blue;options=bold>• </><fg=default;options=bold>{}</>: {}{}".format(
                    title.rstrip("."),
                    description,
                    ",".join("\n    <fg=blue>{}</>".format(link) for link in links),
                ),
                True,
            )

    def _render_trace(self, io, frames):
        from crashtest.frame_collection import FrameCollection

        stack_frames = FrameCollection()
        for frame in frames:
            if (
                self._ignore
                and re.match(self._ignore, frame.filename)
                and not io.is_debug()
            ):
                continue

            stack_frames.append(frame)

        remaining_frames_length = len(stack_frames) - 1
        if io.is_verbose() and remaining_frames_length:
            self._render_line(io, "<fg=yellow>Stack trace</>:", True)
            max_frame_length = len(str(remaining_frames_length))
            frame_collections = stack_frames.compact()
            i = remaining_frames_length
            for collection in frame_collections:
                if collection.is_repeated():
                    if len(collection) > 1:
                        frames_message = "<fg=yellow>{}</> frames".format(
                            len(collection)
                        )
                    else:
                        frames_message = "frame"

                    self._render_line(
                        io,
                        "<fg=blue>{:>{}}</>  Previous {} repeated <fg=blue>{}</> times".format(
                            "...",
                            max_frame_length,
                            frames_message,
                            collection.repetitions,
                        ),
                        True,
                    )

                    i -= len(collection) * collection.repetitions + len(collection)

                for frame in collection:
                    self._render_line(
                        io,
                        "<fg=yellow>{:>{}}</>  <fg=default;options=bold>{}</>:<b>{}</b> in <fg=cyan>{}</>".format(
                            i,
                            max_frame_length,
                            self._get_relative_file_path(frame.filename),
                            frame.lineno,
                            frame.function,
                        ),
                        True,
                    )

                    if io.is_debug():
                        if (frame, 2, 2) not in self._FRAME_SNIPPET_CACHE:
                            code_lines = self._higlighter.code_snippet(
                                frame.file_content, frame.lineno,
                            )

                            self._FRAME_SNIPPET_CACHE[(frame, 2, 2)] = code_lines

                        code_lines = self._FRAME_SNIPPET_CACHE[(frame, 2, 2)]

                        for code_line in code_lines:
                            self._render_line(
                                io,
                                "{:>{}}{}".format(" ", max_frame_length, code_line),
                                indent=1,
                            )
                    else:
                        self._render_line(
                            io,
                            "{:>{}}  {}".format(
                                " ", max_frame_length, frame.line.strip()
                            ),
                        )

                    i -= 1

    def _render_line(
        self, io, line, new_line=False, indent=2
    ):  # type: (IO, str) -> None
        if new_line:
            io.write_line("")

        io.write_line("{}{}".format(indent * " ", line))

    def _get_relative_file_path(self, filepath):
        cwd = os.getcwd()

        if cwd:
            filepath = filepath.replace(cwd + os.path.sep, "")

        home = os.path.expanduser("~")
        if home:
            filepath = filepath.replace(home + os.path.sep, "~" + os.path.sep)

        return filepath

    def _render_traceback(self, io, tb):  # type: (IO, ...) -> None
        frames = []
        while tb:
            frames.append(self._format_traceback_frame(io, tb))

            tb = tb.tb_next

        io.write_line("<b>Traceback (most recent call last):</b>")
        io.write_line("".join(traceback.format_list(frames)))

    def _format_traceback_frame(self, io, tb):  # type: (IO, ...) -> Tuple[Any]
        frame_info = inspect.getframeinfo(tb)
        filename = frame_info.filename
        lineno = frame_info.lineno
        function = frame_info.function
        line = frame_info.code_context[0]

        stripped_line = line.lstrip(" ")
        try:
            tree = ast.parse(stripped_line, mode="exec")
            formatted = self._format_tree(tree, stripped_line, io)
            formatted = (len(line) - len(stripped_line)) * " " + formatted
        except SyntaxError:
            formatted = line

        return (
            io.format("<c1>{}</c1>".format(filename)),
            "<fg=blue;options=bold>{}</>".format(lineno) if not PY2 else lineno,
            "<b>{}</b>".format(function),
            formatted,
        )

    def _format_tree(self, tree, source, io):
        offset = 0
        chunks = []

        nodes = [n for n in ast.walk(tree)]

        displayed_nodes = []

        for node in nodes:
            nodecls = node.__class__
            nodename = nodecls.__name__

            if "col_offset" not in dir(node):
                continue

            if nodecls in self.AST_ELEMENTS["keywords"]:
                displayed_nodes.append((node, nodename.lower(), "keyword"))
            elif nodecls == ast.Name and node.id in self.AST_ELEMENTS["builtins"]:
                displayed_nodes.append((node, node.id, "builtin"))
            elif nodecls == ast.Str:
                displayed_nodes.append((node, "'{}'".format(node.s), "literal"))
            elif nodecls == ast.Num:
                displayed_nodes.append((node, str(node.n), "literal"))

        displayed_nodes.sort(key=lambda elem: elem[0].col_offset)

        for dn in displayed_nodes:
            node = dn[0]
            s = dn[1]
            theme = dn[2]

            begin_col = node.col_offset

            src_chunk = source[offset:begin_col]
            chunks.append(src_chunk)
            chunks.append(io.format("{}{}</>".format(self.THEME[theme], s)))
            offset = begin_col + len(s)

        chunks.append(source[offset:])

        return "".join(chunks)



import re
import time
import threading

from contextlib import contextmanager
from typing import List
from typing import Optional
from typing import Union

from clikit.api.io import IO
from clikit.api.io import Output
from clikit.utils.time import format_time


class ProgressIndicator(object):
    """
    A process indicator.
    """

    NORMAL = " {indicator} {message}"
    NORMAL_NO_ANSI = " {message}"
    VERBOSE = " {indicator} {message} ({elapsed:6s})"
    VERBOSE_NO_ANSI = " {message} ({elapsed:6s})"
    VERY_VERBOSE = " {indicator} {message} ({elapsed:6s})"
    VERY_VERBOSE_NO_ANSI = " {message} ({elapsed:6s})"

    def __init__(
        self, io, fmt=None, interval=100, values=None
    ):  # type: (Union[IO, Output], Optional[str], int, Optional[List[str]]) -> None
        if isinstance(io, IO):
            io = io.error_output

        self._io = io

        if fmt is None:
            fmt = self._determine_best_format()

        self._fmt = fmt

        if values is None:
            values = ["-", "\\", "|", "/"]

        if len(values) < 2:
            raise ValueError(
                "The progress indicator must have at least 2 indicator value characters."
            )

        self._interval = interval
        self._values = values

        self._message = None
        self._update_time = None
        self._started = False
        self._current = 0

        self._auto_running = None
        self._auto_thread = None

        self._start_time = None
        self._last_message_length = 0

    @property
    def message(self):  # type: () -> Optional[str]
        return self._message

    def set_message(self, message):  # type: (Optional[str]) -> None
        self._message = message

        self._display()

    @property
    def current_value(self):  # type: () -> str
        return self._values[self._current % len(self._values)]

    def start(self, message):  # type: (str) -> None
        if self._started:
            raise RuntimeError("Progress indicator already started.")

        self._message = message
        self._started = True
        self._start_time = time.time()
        self._update_time = self._get_current_time_in_milliseconds() + self._interval
        self._current = 0

        self._display()

    def advance(self):  # type: () -> None
        if not self._started:
            raise RuntimeError("Progress indicator has not yet been started.")

        if not self._io.supports_ansi():
            return

        current_time = self._get_current_time_in_milliseconds()
        if current_time < self._update_time:
            return

        self._update_time = current_time + self._interval
        self._current += 1

        self._display()

    def finish(self, message, reset_indicator=False):  # type: (str, bool) -> None
        if not self._started:
            raise RuntimeError("Progress indicator has not yet been started.")

        if self._auto_thread is not None:
            self._auto_running.set()
            self._auto_thread.join()

        self._message = message

        if reset_indicator:
            self._current = 0

        self._display()
        self._io.write_line("")
        self._started = False

    @contextmanager
    def auto(self, start_message, end_message):
        """
        Auto progress.
        """
        self._auto_running = threading.Event()
        self._auto_thread = threading.Thread(target=self._spin)

        self.start(start_message)
        self._auto_thread.start()

        try:
            yield self
        except (Exception, KeyboardInterrupt):
            self._io.write_line("")

            self._auto_running.set()
            self._auto_thread.join()

            raise

        self.finish(end_message, reset_indicator=True)

    def _spin(self):
        while not self._auto_running.is_set():
            self.advance()

            time.sleep(0.1)

    def _display(self):
        if self._io.is_quiet():
            return

        self._overwrite(
            re.sub(
                r"(?i){([a-z\-_]+)(?::([^}]+))?}", self._overwrite_callback, self._fmt
            )
        )

    def _overwrite_callback(self, matches):
        if hasattr(self, "_formatter_{}".format(matches.group(1))):
            text = str(getattr(self, "_formatter_{}".format(matches.group(1)))())
        else:
            text = matches.group(0)

        return text

    def _overwrite(self, message):
        """
        Overwrites a previous message to the output.
        """
        if self._io.supports_ansi():
            self._io.write("\x0D\x1B[2K")
            self._io.write(message)
        else:
            self._io.write_line(message)

    def _determine_best_format(self):
        decorated = self._io.supports_ansi()

        if self._io.is_very_verbose():
            if decorated:
                return self.VERY_VERBOSE

            return self.VERY_VERBOSE_NO_ANSI
        elif self._io.is_verbose():
            if decorated:
                return self.VERY_VERBOSE

            return self.VERBOSE_NO_ANSI

        if decorated:
            return self.NORMAL

        return self.NORMAL_NO_ANSI

    def _get_current_time_in_milliseconds(self):
        return round(time.time() * 1000)

    def _formatter_indicator(self):
        return self.current_value

    def _formatter_message(self):
        return self.message

    def _formatter_elapsed(self):
        return format_time(time.time() - self._start_time)



from .args import Args
from .format.args_format import ArgsFormat
from .raw_args import RawArgs


class ArgsParser(object):
    """
    Parses raw console arguments and returns the parsed arguments.
    """

    def parse(
        self, args, fmt, lenient=False
    ):  # type: (RawArgs, ArgsFormat, bool) -> Args
        raise NotImplementedError()



from .api.config.application_config import ApplicationConfig
from .config.default_application_config import DefaultApplicationConfig
from .console_application import ConsoleApplication

__version__ = "0.6.1"



# -*- coding: utf-8 -*-
from __future__ import division

import time
import re
import math

from typing import Union

from clikit.api.io import IO
from clikit.api.io.flags import DEBUG
from clikit.api.io.flags import VERBOSE
from clikit.api.io.flags import VERY_VERBOSE
from clikit.api.io.section_output import SectionOutput
from clikit.api.io.output import Output
from clikit.utils.terminal import Terminal
from clikit.utils.time import format_time


class ProgressBar(object):
    """
    The ProgressBar provides helpers to display progress output.
    """

    # Options
    bar_width = 28
    bar_char = None
    empty_bar_char = "-"
    progress_char = ">"
    redraw_freq = 1

    formats = {
        "normal": " %current%/%max% [%bar%] %percent:3s%%",
        "normal_nomax": " %current% [%bar%]",
        "verbose": " %current%/%max% [%bar%] %percent:3s%% %elapsed:-6s%",
        "verbose_nomax": " %current% [%bar%] %elapsed:6s%",
        "very_verbose": " %current%/%max% [%bar%] %percent:3s%% %elapsed:6s%/%estimated:-6s%",
        "very_verbose_nomax": " %current% [%bar%] %elapsed:6s%",
        "debug": " %current%/%max% [%bar%] %percent:3s%% %elapsed:6s%/%estimated:-6s%",
        "debug_nomax": " %current% [%bar%] %elapsed:6s%",
    }

    def __init__(
        self, io, max=0, min_seconds_between_redraws=0.1
    ):  # type: (Union[IO, Output], int, float) -> None
        """
        Constructor.
        """
        # If we have an IO, ensure we write to the error output
        if isinstance(io, IO):
            io = io.error_output

        self._io = io
        self._terminal = Terminal()
        self._max = 0
        self._step_width = None
        self._set_max_steps(max)
        self._step = 0
        self._percent = 0.0
        self._format = None
        self._internal_format = None
        self._format_line_count = 0
        self._last_messages_length = 0
        self._should_overwrite = True
        self._min_seconds_between_redraws = 0
        self._max_seconds_between_redraws = 1
        self._write_count = 0

        if min_seconds_between_redraws > 0:
            self.redraw_freq = None
            self._min_seconds_between_redraws = min_seconds_between_redraws

        if not self._io.supports_ansi():
            # Disable overwrite when output does not support ANSI codes.
            self._should_overwrite = False

            # Set a reasonable redraw frequency so output isn't flooded
            self.redraw_freq = None

        self._messages = {}

        self._start_time = time.time()
        self._last_write_time = 0

    def set_message(self, message, name="message"):
        self._messages[name] = message

    def get_message(self, name="message"):
        return self._messages[name]

    def get_start_time(self):
        return self._start_time

    def get_max_steps(self):
        return self._max

    def get_progress(self):
        return self._step

    def get_progress_percent(self):
        return self._percent

    def set_bar_character(self, character):
        self.bar_char = character

        return self

    def get_bar_character(self):
        if self.bar_char is None:
            if self._max:
                return "="

            return self.empty_bar_char

        return self.bar_char

    def get_bar_width(self):
        return self.bar_width

    def set_bar_width(self, width):
        self.bar_width = width

        return self

    def get_empty_bar_character(self):
        return self.empty_bar_char

    def set_empty_bar_character(self, character):
        self.empty_bar_char = character

        return self

    def get_progress_character(self):
        return self.progress_char

    def set_progress_character(self, character):
        self.progress_char = character

        return self

    def set_format(self, fmt):
        self._format = None
        self._internal_format = fmt

    def set_redraw_frequency(self, freq):
        if self.redraw_freq is not None:
            self.redraw_freq = max(freq, 1)

    def min_seconds_between_redraws(self, freq):  # type:  (float) -> None
        if freq > 0:
            self.redraw_freq = None
            self._min_seconds_between_redraws = freq

    def max_seconds_between_redraws(self, freq):  # type: (float) -> None
        self._max_seconds_between_redraws = freq

    def start(self, max=None):
        """
        Start the progress output.
        """
        self._start_time = time.time()
        self._step = 0
        self._percent = 0.0

        if max is not None:
            self._set_max_steps(max)

        self.display()

    def advance(self, step=1):
        """
        Advances the progress output X steps.
        """
        self.set_progress(self._step + step)

    def set_progress(self, step):
        """
        Sets the current progress.

        :param step: The current progress
        :type step: int
        """
        if self._max and step > self._max:
            self._max = step
        elif step < 0:
            step = 0

        redraw_freq = (
            (self._max or 10) / 10 if self.redraw_freq is None else self.redraw_freq
        )
        prev_period = int(self._step / redraw_freq)
        curr_period = int(step / redraw_freq)

        self._step = step

        if self._max:
            self._percent = self._step / self._max
        else:
            self._percent = 0.0

        time_interval = time.time() - self._last_write_time

        # Draw regardless of other limits
        if step == self._max:
            self.display()

            return

        # Throttling
        if time_interval < self._min_seconds_between_redraws:
            return

        # Draw each step period, but not too late
        if (
            prev_period != curr_period
            or time_interval >= self._max_seconds_between_redraws
        ):
            self.display()

    def finish(self):
        """
        Finish the progress output.
        """
        if not self._max:
            self._max = self._step

        if self._step == self._max and not self._should_overwrite:
            return

        self.set_progress(self._max)

    def display(self):
        """
        Output the current progress string.
        """
        if self._io.is_quiet():
            return

        if self._format is None:
            self._set_real_format(
                self._internal_format or self._determine_best_format()
            )

        self._overwrite(
            re.sub(
                r"(?i)%([a-z\-_]+)(?::([^%]+))?%",
                self._overwrite_callback,
                self._format,
            )
        )

    def _overwrite_callback(self, matches):
        if hasattr(self, "_formatter_{}".format(matches.group(1))):
            text = str(getattr(self, "_formatter_{}".format(matches.group(1)))())
        elif matches.group(1) in self._messages:
            text = self._messages[matches.group(1)]
        else:
            return matches.group(0)

        if matches.group(2):
            if matches.group(2).startswith("-"):
                text = text.ljust(int(matches.group(2).lstrip("-").rstrip("s")))
            else:
                text = text.rjust(int(matches.group(2).rstrip("s")))

        return text

    def clear(self):
        """
        Removes the progress bar from the current line.

        This is useful if you wish to write some output
        while a progress bar is running.
        Call display() to show the progress bar again.
        """
        if not self._should_overwrite:
            return

        if self._format is None:
            self._set_real_format(
                self._internal_format or self._determine_best_format()
            )

        self._overwrite("\n" * self._format_line_count)

    def _set_real_format(self, fmt):
        """
        Sets the progress bar format.
        """
        # try to use the _nomax variant if available
        if not self._max and fmt + "_nomax" in self.formats:
            self._format = self.formats[fmt + "_nomax"]
        elif fmt in self.formats:
            self._format = self.formats[fmt]
        else:
            self._format = fmt

        self._format_line_count = self._format.count("\n")

    def _set_max_steps(self, mx):
        """
        Sets the progress bar maximal steps.
        """
        self._max = max(0, mx)

        if self._max:
            self._step_width = len(str(self._max))
        else:
            self._step_width = 4

    def _overwrite(self, message):
        """
        Overwrites a previous message to the output.
        """
        lines = message.split("\n")

        # Append whitespace to match the line's length
        if self._last_messages_length is not None:
            for i, line in enumerate(lines):
                if self._last_messages_length > len(self._io.remove_format(line)):
                    lines[i] = line.ljust(self._last_messages_length, "\x20")

        if self._should_overwrite:
            if isinstance(self._io, SectionOutput):
                lines_to_clear = (
                    int(math.floor(len(lines) / self._terminal.width))
                    + self._format_line_count
                    + 1
                )
                self._io.clear(lines_to_clear)
            else:
                # move back to the beginning of the progress bar before redrawing it
                self._io.write("\x0D")

                if self._format_line_count:
                    self._io.write("\033[{}A".format(self._format_line_count))
        elif self._step > 0:
            # move to new line
            self._io.write_line("")

        self._io.write("\n".join(lines))
        self._io.flush()

        self._last_messages_length = 0

        for line in lines:
            length = len(self._io.remove_format(line))
            if length > self._last_messages_length:
                self._last_messages_length = length

        self._last_write_time = time.time()
        self._write_count += 1

    def _determine_best_format(self):
        verbosity = self._io.verbosity

        if verbosity == VERBOSE:
            if self._max:
                return "verbose"

            return "verbose_nomax"
        elif verbosity == VERY_VERBOSE:
            if self._max:
                return "very_verbose"

            return "very_verbose_nomax"
        elif verbosity == DEBUG:
            if self._max:
                return "debug"

            return "debug_nomax"

        if self._max:
            return "normal"

        return "normal_nomax"

    @property
    def bar_offset(self):  # type: () -> int
        if self._max:
            return math.floor(self._percent * self.bar_width)
        else:
            if self.redraw_freq is None:
                return math.floor(
                    (min(5, self.get_bar_width() / 15) * self._write_count)
                    % self.bar_width
                )

            return math.floor(self._step % self.bar_width)

    def _formatter_bar(self):
        complete_bars = self.bar_offset

        display = self.get_bar_character() * int(complete_bars)

        if complete_bars < self.bar_width:
            empty_bars = (
                self.bar_width
                - complete_bars
                - len(self._io.remove_format(self.progress_char))
            )
            display += self.progress_char + self.empty_bar_char * int(empty_bars)

        return display

    def _formatter_elapsed(self):
        return format_time(time.time() - self._start_time)

    def _formatter_remaining(self):
        if not self._max:
            raise RuntimeError(
                "Unable to display the remaining time "
                "if the maximum number of steps is not set."
            )

        if not self._step:
            remaining = 0
        else:
            remaining = round(
                (time.time() - self._start_time) / self._step * (self._max - self._max)
            )

        return format_time(remaining)

    def _formatter_estimated(self):
        if not self._max:
            raise RuntimeError(
                "Unable to display the estimated time "
                "if the maximum number of steps is not set."
            )

        if not self._step:
            estimated = 0
        else:
            estimated = round((time.time() - self._start_time) / self._step * self._max)

        return estimated

    def _formatter_current(self):
        return str(self._step).rjust(self._step_width, " ")

    def _formatter_max(self):
        return self._max

    def _formatter_percent(self):
        return int(math.floor(self._percent * 100))



import re

from typing import Any
from typing import Optional

from ._compat import basestring


def parse_string(value, nullable=True):  # type: (Any, bool) -> Optional[str]
    if nullable and (value is None or value == "null"):
        return

    if value is None:
        return "null"

    if isinstance(value, bool):
        return str(value).lower()

    return str(value)


def parse_boolean(value, nullable=True):  # type: (Any, bool) -> Optional[bool]
    if nullable and (value is None or value == "null"):
        return

    if isinstance(value, bool):
        return value

    if isinstance(value, int):
        value = str(value)

    if isinstance(value, basestring):
        if not value:
            return False

        if value in {"false", "0", "no", "off"}:
            return False

        if value in {"true", "1", "yes", "on"}:
            return True

    raise ValueError('The value "{}" cannot be parsed as boolean.'.format(value))


def parse_int(value, nullable=True):  # type: (Any, bool) -> Optional[int]
    if nullable and (value is None or value == "null"):
        return

    try:
        return int(value)
    except ValueError:
        raise ValueError('The value "{}" cannot be parsed as integer.'.format(value))


def parse_float(value, nullable=True):  # type: (Any, bool) -> Optional[float]
    if nullable and (value is None or value == "null"):
        return

    try:
        return float(value)
    except ValueError:
        raise ValueError('The value "{}" cannot be parsed as float.'.format(value))


def get_string_length(
    string, formatter=None
):  # type: (str, Optional[Formatter]) -> int
    if formatter is not None:
        string = formatter.remove_format(string)

    return len(string)


def get_max_word_length(
    string, formatter=None
):  # type: (str, Optional[Formatter]) -> int
    if formatter is not None:
        string = formatter.remove_format(string)

    max_length = 0
    words = re.split("\s+", string)

    for word in words:
        max_length = max(max_length, get_string_length(word))

    return max_length


def get_max_line_length(
    string, formatter=None
):  # type: (str, Optional[Formatter]) -> int
    if formatter is not None:
        string = formatter.remove_format(string)

    max_length = 0
    words = re.split("\n", string)

    for word in words:
        max_length = max(max_length, get_string_length(word))

    return max_length



from clikit.api.args.format.args_format import ArgsFormat
from clikit.api.args.raw_args import RawArgs
from clikit.api.command.command_collection import CommandCollection
from clikit.api.config.application_config import ApplicationConfig
from clikit.api.io import InputStream
from clikit.api.io import OutputStream
from clikit.api.resolver.resolved_command import ResolvedCommand


class Application:
    """
    A console application.
    """

    @property
    def config(self):  # type: () -> ApplicationConfig
        raise NotImplementedError()

    @property
    def global_args_format(self):  # type: () -> ArgsFormat
        raise NotImplementedError()

    def get_command(self, name):  # type: (str) -> Command
        raise NotImplementedError()

    @property
    def commands(self):  # type: () -> CommandCollection
        raise NotImplementedError()

    def has_command(self, name):  # type: (str) -> bool
        raise NotImplementedError()

    def has_commands(self):  # type: () -> bool
        raise NotImplementedError()

    @property
    def named_commands(self):  # type: () -> CommandCollection
        raise NotImplementedError()

    def has_named_commands(self):  # type: () -> bool
        raise NotImplementedError()

    @property
    def default_commands(self):  # type: () -> CommandCollection
        raise NotImplementedError()

    def has_default_commands(self):  # type: () -> bool
        raise NotImplementedError()

    def resolve_command(self, args):  # type: (RawArgs) -> ResolvedCommand
        raise NotImplementedError()

    def run(
        self, args=None, input_stream=None, output_stream=None, error_stream=None
    ):  # type: (RawArgs, InputStream, OutputStream, OutputStream) -> int
        raise NotImplementedError()



from pastel.style import Style as PastelStyle

from clikit.api.formatter import Style


class StyleConverter(object):
    """
    Converts a TTY Style instance to a Pastel Style instance.
    """

    @classmethod
    def convert(cls, style):  # type: (Style) -> PastelStyle
        options = []

        if style.is_bold():
            options.append("bold")

        if style.is_italic():
            options.append("italic")

        if style.is_dark():
            options.append("dark")

        if style.is_underlined():
            options.append("underline")

        if style.is_blinking():
            options.append("blink")

        if style.is_inverse():
            options.append("reverse")

        if style.is_hidden():
            options.append("conceal")

        return PastelStyle(style.foreground_color, style.background_color, options)



from clikit.api.args import RawArgs
from clikit.api.resolver import ResolvedCommand

from .event import Event


class PreResolveEvent(Event):
    """
    Dispatched before the console arguments are resolved to a command.

    Add a listener for this event to customize the command used for the given
    console arguments.
    """

    def __init__(self, raw_args, application):  # type: (RawArgs, Application) -> None
        super(PreResolveEvent, self).__init__()

        self._raw_args = raw_args
        self._application = application
        self._resolved_command = None

    @property
    def raw_args(self):  # type: () -> RawArgs
        return self._raw_args

    @property
    def application(self):  # type: () -> Application
        return self._application

    @property
    def resolved_command(self):  # type: () -> ResolvedCommand
        return self._resolved_command

    def set_resolved_command(
        self, resolved_command=None
    ):  # type: (ResolvedCommand) -> None
        self._resolved_command = resolved_command



from __future__ import unicode_literals

import re
import textwrap

from clikit.api.io import IO
from clikit.ui import Component


class LabeledParagraph(Component):
    """
    A paragraph with a label on its left.
    """

    def __init__(
        self, label, text, padding=2, aligned=True
    ):  # type: (str, str, int, bool) -> None
        self._label = label
        self._text = text
        self._padding = padding
        self._aligned = aligned
        self._alignment = None

    @property
    def label(self):  # type: () -> str
        return self._label

    @property
    def text(self):  # type: () -> str
        return self._text

    @property
    def padding(self):  # type: () -> int
        return self._padding

    def is_aligned(self):  # type: () -> bool
        return self._aligned

    def set_alignment(self, alignment):  # type: (LabelAlignment) -> None
        self._alignment = alignment

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        line_prefix = " " * indentation
        visible_label = io.remove_format(self._label)
        style_tag_length = len(self._label) - len(visible_label)

        if self._aligned and self._alignment:
            text_offset = self._alignment.text_offset - indentation
        else:
            text_offset = 0

        text_offset = max(text_offset, len(visible_label) + self._padding)
        text_prefix = " " * text_offset

        # 1 trailing space
        text_width = io.terminal_dimensions.width - 1 - text_offset - indentation
        text = re.sub(
            r"\n(?!\n)",
            "\n" + line_prefix + text_prefix,
            "\n".join(textwrap.wrap(self._text, text_width)),
        )

        # Add the total length of the style tags ("<b>", ...)
        label_width = text_offset + style_tag_length

        io.write(
            "{}{:<{}}{}".format(
                line_prefix, self._label, label_width, text.rstrip()
            ).rstrip()
            + "\n"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['clikit',
 'clikit.adapter',
 'clikit.api',
 'clikit.api.application',
 'clikit.api.args',
 'clikit.api.args.format',
 'clikit.api.command',
 'clikit.api.config',
 'clikit.api.event',
 'clikit.api.formatter',
 'clikit.api.io',
 'clikit.api.resolver',
 'clikit.args',
 'clikit.args.inputs',
 'clikit.config',
 'clikit.formatter',
 'clikit.handler',
 'clikit.handler.help',
 'clikit.io',
 'clikit.io.input_stream',
 'clikit.io.output_stream',
 'clikit.resolver',
 'clikit.ui',
 'clikit.ui.alignment',
 'clikit.ui.components',
 'clikit.ui.help',
 'clikit.ui.layout',
 'clikit.ui.style',
 'clikit.utils']

package_data = \
{'': ['*']}

install_requires = \
['pastel>=0.2.0,<0.3.0', 'pylev>=1.3,<2.0']

extras_require = \
{':python_version >= "2.7" and python_version < "2.8"': ['enum34>=1.1,<2.0'],
 ':python_version >= "2.7" and python_version < "2.8" or python_version >= "3.4" and python_version < "3.5"': ['typing>=3.6,<4.0'],
 ':python_version >= "3.5.0" and python_version < "3.5.4"': ['typing-extensions>=3.6,<4.0'],
 ':python_version >= "3.6" and python_version < "4.0"': ['crashtest>=0.3.0,<0.4.0']}

setup_kwargs = {
    'name': 'clikit',
    'version': '0.6.1',
    'description': 'CliKit is a group of utilities to build beautiful and testable command line interfaces.',
    'long_description': '# CliKit\n\nCliKit is a group of utilities to build beautiful and testable command line interfaces.\n\nThis is at the core of [Cleo](https://github.com/sdispater/cleo).\n',
    'author': 'Sébastien Eustace',
    'author_email': 'sebastien@eustace.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdispater/clikit',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)



import re
import textwrap

from clikit.api.io import IO
from clikit.ui import Component


class Paragraph(Component):
    """
    A paragraph of text.

    The paragraph is wrapped into the dimensions of the output.
    """

    def __init__(self, text):  # type: (str) -> None
        self._text = text

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        line_prefix = " " * indentation
        text_width = io.terminal_dimensions.width - 1 - indentation

        text = re.sub(
            r"\n(?!\n)",
            "\n" + line_prefix,
            "\n".join(textwrap.wrap(self._text, text_width)),
        )

        io.write(line_prefix + text.rstrip() + "\n")
